"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isAdmin = exports.getWADExecutablePath = exports.setupWAD = exports.downloadWAD = void 0;
const lodash_1 = __importDefault(require("lodash"));
const support_1 = require("appium/support");
const path_1 = __importDefault(require("path"));
const teen_process_1 = require("teen_process");
const logger_1 = __importDefault(require("./logger"));
const registry_1 = require("./registry");
const utils_1 = require("./utils");
// https://github.com/microsoft/WinAppDriver/releases
const WAD_VER = '1.2.99';
const WAD_DOWNLOAD_MD5 = Object.freeze({
    x32: '23745e6ed373bc969ff7c4493e32756a',
    x64: '2923fc539f389d47754a7521ee50108e',
    arm64: 'b9af4222a3fb0d688ecfbf605d1c4500',
});
const ARCH_MAPPING = Object.freeze({ x32: 'x86', x64: 'x64', arm64: 'arm64' });
const WAD_DOWNLOAD_TIMEOUT_MS = 60000;
const POSSIBLE_WAD_INSTALL_ROOTS = [
    process.env['ProgramFiles(x86)'],
    process.env.ProgramFiles,
    `${process.env.SystemDrive || 'C:'}\\\\Program Files`,
];
const WAD_EXE_NAME = 'WinAppDriver.exe';
const UNINSTALL_REG_ROOT = 'HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall';
const REG_ENTRY_VALUE = 'Windows Application Driver';
const REG_ENTRY_KEY = 'DisplayName';
const REG_ENTRY_TYPE = 'REG_SZ';
const INST_LOCATION_SCRIPT_BY_GUID = (guid) => `
Set installer = CreateObject("WindowsInstaller.Installer")
Set session = installer.OpenProduct("${guid}")
session.DoAction("CostInitialize")
session.DoAction("CostFinalize")
WScript.Echo session.Property("INSTALLFOLDER")
`.replace(/\n/g, '\r\n');
function generateWadDownloadLink() {
    const wadArch = ARCH_MAPPING[process.arch];
    if (!wadArch) {
        throw new Error(`System architecture '${process.arch}' is not supported by Windows Application Driver. ` +
            `The only supported architectures are: ${lodash_1.default.keys(ARCH_MAPPING)}`);
    }
    return `https://github.com/Microsoft/WinAppDriver` +
        `/releases/download/v${WAD_VER}/WindowsApplicationDriver-${WAD_VER}-win-${wadArch}.exe`;
}
async function fetchMsiInstallLocation(installerGuid) {
    const tmpRoot = await support_1.tempDir.openDir();
    const scriptPath = path_1.default.join(tmpRoot, 'get_wad_inst_location.vbs');
    try {
        await support_1.fs.writeFile(scriptPath, INST_LOCATION_SCRIPT_BY_GUID(installerGuid), 'latin1');
        const { stdout } = await (0, utils_1.shellExec)('cscript.exe', ['/Nologo', scriptPath]);
        return lodash_1.default.trim(stdout);
    }
    finally {
        await support_1.fs.rimraf(tmpRoot);
    }
}
class WADNotFoundError extends Error {
}
const getWADExecutablePath = lodash_1.default.memoize(async function getWADInstallPath() {
    const wadPath = process.env.APPIUM_WAD_PATH ?? '';
    if (await support_1.fs.exists(wadPath)) {
        logger_1.default.debug(`Loaded WinAppDriver path from the APPIUM_WAD_PATH environment variable: ${wadPath}`);
        return wadPath;
    }
    // TODO: WAD installer should write the full path to it into the system registry
    const pathCandidates = POSSIBLE_WAD_INSTALL_ROOTS
        // remove unset env variables
        .filter(Boolean)
        // construct full path
        // @ts-ignore The above filter does the job
        .map((root) => path_1.default.resolve(root, REG_ENTRY_VALUE, WAD_EXE_NAME));
    for (const result of pathCandidates) {
        if (await support_1.fs.exists(result)) {
            return result;
        }
    }
    logger_1.default.debug('Did not detect WAD executable at any of the default install locations');
    logger_1.default.debug('Checking the system registry for the corresponding MSI entry');
    try {
        const uninstallEntries = await (0, registry_1.queryRegistry)(UNINSTALL_REG_ROOT);
        const wadEntry = uninstallEntries.find(({ key, type, value }) => key === REG_ENTRY_KEY && value === REG_ENTRY_VALUE && type === REG_ENTRY_TYPE);
        if (wadEntry) {
            logger_1.default.debug(`Found MSI entry: ${JSON.stringify(wadEntry)}`);
            const installerGuid = lodash_1.default.last(wadEntry.root.split('\\'));
            // WAD MSI installer leaves InstallLocation registry value empty,
            // so we need to be hacky here
            const result = path_1.default.join(await fetchMsiInstallLocation(installerGuid), WAD_EXE_NAME);
            logger_1.default.debug(`Checking if WAD exists at '${result}'`);
            if (await support_1.fs.exists(result)) {
                return result;
            }
            logger_1.default.debug(result);
        }
        else {
            logger_1.default.debug('No WAD MSI entries have been found');
        }
    }
    catch (e) {
        if (e.stderr) {
            logger_1.default.debug(e.stderr);
        }
        logger_1.default.debug(e.stack);
    }
    throw new WADNotFoundError(`${WAD_EXE_NAME} has not been found in any of these ` +
        `locations: ${pathCandidates}. Is it installed?`);
});
exports.getWADExecutablePath = getWADExecutablePath;
async function downloadWAD() {
    const downloadLink = generateWadDownloadLink();
    const installerPath = path_1.default.resolve(await support_1.tempDir.staticDir(), `wad_installer_${WAD_VER}_${support_1.util.uuidV4()}.exe`);
    logger_1.default.info(`Downloading ${downloadLink} to '${installerPath}'`);
    await support_1.net.downloadFile(downloadLink, installerPath, { timeout: WAD_DOWNLOAD_TIMEOUT_MS });
    const downloadedMd5 = await support_1.fs.md5(installerPath);
    const expectedMd5 = WAD_DOWNLOAD_MD5[process.arch];
    if (downloadedMd5 !== expectedMd5) {
        await support_1.fs.rimraf(installerPath);
        throw new Error(`Installer executable checksum validation error: expected ${expectedMd5} but got ${downloadedMd5}`);
    }
    return installerPath;
}
exports.downloadWAD = downloadWAD;
const isAdmin = async function isAdmin() {
    try {
        await (0, teen_process_1.exec)('fsutil.exe', ['dirty', 'query', process.env.SystemDrive || 'C:']);
        return true;
    }
    catch (ign) {
        return false;
    }
};
exports.isAdmin = isAdmin;
async function setupWAD() {
    if (!support_1.system.isWindows()) {
        throw new Error(`Can only download WinAppDriver on Windows!`);
    }
    try {
        return await getWADExecutablePath();
    }
    catch (e) {
        if (!(e instanceof WADNotFoundError)) {
            throw e;
        }
        logger_1.default.info(`WinAppDriver doesn't exist, setting up`);
    }
    const installerPath = await downloadWAD();
    logger_1.default.info(`Running WinAppDriver v${WAD_VER} installer`);
    try {
        await (0, utils_1.shellExec)(installerPath, ['/install', '/quiet', '/norestart']);
    }
    catch (e) {
        /** @type {import('node:child_process').ExecException} */
        const error = e;
        throw new Error(`WinAppDriver cannot be installed: ${error.message}. Exit code: ${error.code}`);
    }
    finally {
        await support_1.fs.rimraf(installerPath);
    }
}
exports.setupWAD = setupWAD;
exports.default = setupWAD;
//# sourceMappingURL=installer.js.map